/*
 * Support for executing code in new tasks and threads.
 *
 * Copyright (C) 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>
#include <l4/sys/factory.h>
#include <l4/sys/task.h>
#include <l4/sys/thread.h>
#include <l4/util/bitops.h> /* l4util_log2 */
#include <l4/util/util.h>   /* l4_touch_rw */

#include <l4/re/c/mem_alloc.h>
#include <l4/re/c/rm.h>
#include <l4/re/c/util/cap_alloc.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pthread-l4.h>
#include <pthread.h>



/* Pager properties. */

static l4_umword_t pager_id;
static char *buf;
static unsigned int buf_log2size;
static l4_addr_t buf_start;



/* Handle an exception, this being here just in case we get one. */

static void handle_exception(l4_msgtag_t)
{
    l4_exc_regs_t *regs = (l4_exc_regs_t *) l4_utcb_mr()->mr;

    printf("exception(...) -> pfa = 0x%lx, pc = 0x%lx\n", l4_utcb_exc_pfa(regs), l4_utcb_exc_pc(regs));

    l4_ipc_send(L4_INVALID_CAP | L4_SYSF_REPLY, l4_utcb(), l4_msgtag(L4_EOK, 0, 0, 0), L4_IPC_NEVER);
}

/* Handle a page fault in the configured address range. */

static void handle_page_fault(l4_msgtag_t)
{
    l4_umword_t pfa = l4_utcb_mr()->mr[0];
    l4_umword_t pc = l4_utcb_mr()->mr[1];

    /* Obtain the actual fault address and flags. */

    l4_umword_t addr = pfa & ~7UL, flags = pfa & 7;

    printf("page_fault(0x%lx, 0x%lx) -> 0x%lx (0x%lx)...\n", pfa, pc, addr, flags);

    /* Check the configured range. */

    if ((addr >= buf_start) && (addr < buf_start + (1UL << buf_log2size)))
    {
        l4_addr_t buf_addr = (l4_addr_t) buf;

        /* Construct a flexpage item and define a "send base" to resolve the fault. */

        l4_fpage_t fpage = l4_fpage(buf_addr, buf_log2size, L4_FPAGE_RX);
        l4_umword_t snd_base = buf_start;

        printf("-> l4_fpage(0x%lx, %d, 0x%x) @ %lx\n",
               l4_fpage_memaddr(fpage), l4_fpage_size(fpage), l4_fpage_rights(fpage),
               snd_base);

        l4_utcb_mr()->mr[0] = l4_map_control(snd_base, L4_FPAGE_CACHEABLE, 0);
        l4_utcb_mr()->mr[1] = fpage.raw;

        l4_ipc_send(L4_INVALID_CAP | L4_SYSF_REPLY, l4_utcb(), l4_msgtag(L4_EOK, 0, 1, 0), L4_IPC_NEVER);
        return;
    }

    l4_ipc_send(L4_INVALID_CAP | L4_SYSF_REPLY, l4_utcb(), l4_msgtag(-L4_ENOMEM, 0, 0, 0), L4_IPC_NEVER);
}

/* Dispatch to the appropriate handler. */

static void handle_message(l4_msgtag_t tag)
{
  switch (l4_msgtag_label(tag))
  {
    case L4_PROTO_EXCEPTION:
      handle_exception(tag);
      break;

    case L4_PROTO_PAGE_FAULT:
      handle_page_fault(tag);
      break;

    default:
      l4_ipc_send(L4_INVALID_CAP | L4_SYSF_REPLY, l4_utcb(), l4_msgtag(-L4_EBADPROTO, 0, 0, 0), L4_IPC_NEVER);
      break;
  }
}

/* Reserve a capability slot and associate it with an IPC gate for the pager. */

static long init_pager(l4_cap_idx_t *ref, l4_cap_idx_t thread)
{ 
  *ref = l4re_util_cap_alloc();
  
  if (l4_is_invalid_cap(*ref))
    return -L4_ENOENT;
  
  long err = l4_error(l4_factory_create_gate(l4re_env()->factory, *ref, thread, pager_id));
  
  if (err)
    l4re_util_cap_free_um(*ref);

  return err;
}

/* Receive IPC messages for exception and page fault handling. */

static void *start_pager(void *)
{
  while (1)
  {
    l4_umword_t label;
    l4_msgtag_t tag = l4_ipc_wait(l4_utcb(), &label, L4_IPC_NEVER);

    /* Clear lower label bits. */

    label = label & ~3UL;

    /* Ignore erroneous messages. */

    if (l4_ipc_error(tag, l4_utcb()))
      continue;

    if (label == pager_id)
      handle_message(tag);
  }

  return NULL;
}



/* UTCB properties.
   See: moe/server/src/loader.cc */

enum ipc_exec_utcb
{
  Default_max_threads = 16,
#ifdef ARCH_mips
  Utcb_area_start = 0x73000000,
#else
  Utcb_area_start = 0xb3000000,
#endif
};



/* Utilities to define appropriate region size indicators. */

static int find_region_exp(size_t size)
{
  int exp = l4util_log2(size);

  if ((1UL << exp) < size)
    return exp + 1;
  else
    return exp;
}

static size_t find_region_size(size_t size)
{
  return 1 << find_region_exp(size);
}



/* Allocate a capability slot, allocate a dataspace, and attach the dataspace to
   the task. */

static long allocate(unsigned long size, void **addr, l4re_ds_t *ds)
{
  *ds = l4re_util_cap_alloc();

  if (l4_is_invalid_cap(*ds))
    return -L4_ENOENT;

  if (l4re_ma_alloc_align(size, *ds, 0, 0))
  {
    l4re_util_cap_free_um(*ds);
    return -L4_ENOMEM;
  }

  if (l4re_rm_attach(addr, size, L4RE_RM_F_SEARCH_ADDR | L4RE_RM_F_RW, *ds, 0,
                     L4_PAGESHIFT))
  {
    l4re_util_cap_free_um(*ds);
    return -L4_ENOMEM;
  }

  return L4_EOK;
}



/* Main program, creating a new task and starting a new thread, and running a
   pager thread to handle exceptions and page faults. */

int main(int argc, char *argv[])
{
  long err;

  if (argc < 2)
  {
    printf("Need a program to run.\n");
    return 1;
  }

  /* Allocate capabilities for the task and thread. */

  l4_cap_idx_t task, thread;

  task = l4re_util_cap_alloc();

  if (l4_is_invalid_cap(task))
  {
    printf("Could not allocate task.\n");
    return 1;
  }

  thread = l4re_util_cap_alloc();

  if (l4_is_invalid_cap(thread))
  {
    printf("Could not allocate thread.\n");
    return 1;
  }

  /* Obtain the payload for loading. */

  FILE *file = fopen(argv[1], "r");

  if (file == NULL)
  {
    printf("Could not read file: %s\n", argv[1]);
    return 1;
  }

  /* Read the payload in its entirety into allocated memory. */

  size_t nread;
  size_t max_file_size = 0x40000;
  size_t region_size = find_region_size(max_file_size);

  if (posix_memalign((void **) &buf, region_size, region_size))
  {
    printf("Could not reserve memory.\n");
    return 1;
  }

  printf("Reading from file into %p.\n", buf);

  nread = fread(buf, sizeof(char), max_file_size, file);

  printf("Read %ld from file.\n", nread);

  /* Test the contents of the payload at the entry point locations. */

  if (memcmp(buf + 0xae3, "\x31\xed", 2))
  {
    printf("Did not find expected instructions at start.\n");
    return 1;
  }

  /* UTCB location and size. */

  l4_addr_t utcb_start = Utcb_area_start;
  int utcb_log2size = find_region_exp(Default_max_threads * L4_UTCB_OFFSET);
  l4re_ds_t utcb_ds;

  /* Round up to at least one page. */

  if (utcb_log2size < L4_PAGESHIFT)
    utcb_log2size = L4_PAGESHIFT;

  size_t utcb_size = 1UL << utcb_log2size;

  /* Allocate a dataspace for the UTCB. */

  err = allocate(utcb_size, (void **) &utcb_start, &utcb_ds);

  if (err)
  {
    printf("Could not allocate UTCB: %s\n", l4sys_errtostr(err));
    return 1;
  }

  l4_touch_rw((const void *) utcb_start, utcb_size);

  /* KIP allocation, presumed to be the same in all tasks. */

  l4_addr_t kip_start = (l4_addr_t) l4re_kip();

  printf("KIP at %lx.\n", kip_start);

  /* Create a new task and thread. */

  l4_fpage_t utcb_fpage = l4_fpage(utcb_start, utcb_log2size, 0);

  err = l4_error(l4_factory_create_task(l4re_env()->factory, task, utcb_fpage));

  if (err)
  {
    printf("Could not create task.\n");
    return 1;
  }

  err = l4_error(l4_factory_create_thread(l4re_env()->factory, thread));

  if (err)
  {
    printf("Could not create thread.\n");
    return 1;
  }

  /* Set pager properties. */

  pager_id = 0x1234;
  buf_log2size = find_region_exp(nread);
  buf_start = 0x1000000;

  /* Start the pager. */

  pthread_t pager_thread;
  pthread_attr_t attr;

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

  err = pthread_create(&pager_thread, &attr, start_pager, NULL);

  if (err)
  {
    printf("Could not start pager thread.\n");
    return 1;
  }

  l4_cap_idx_t server;

  err = init_pager(&server, pthread_l4_cap(pager_thread));

  if (err)
  {
    printf("Could not start pager.\n");
    return 1;
  }

  /* Map the pager capability into the region manager/mapper slot. */

  err = l4_error(l4_task_map(task, L4RE_THIS_TASK_CAP,
                             l4_obj_fpage(server, 0, L4_CAP_FPAGE_RWS),
                             l4_map_obj_control(0x10 << L4_CAP_SHIFT, L4_MAP_ITEM_MAP)));

  if (err)
  {
    printf("Could not map pager capability into task.\n");
    return 1;
  }

  /* Map the KIP into the task. */

  err = l4_error(l4_task_map(task, L4RE_THIS_TASK_CAP,
                             l4_fpage(kip_start, L4_PAGESHIFT, L4_FPAGE_RX),
                             kip_start));

  if (err)
  {
    printf("Could not map KIP into task.\n");
    return 1;
  }

  err = l4_error(l4_task_map(task, L4RE_THIS_TASK_CAP,
                             l4_obj_fpage(task, 0, L4_CAP_FPAGE_RWS),
                             l4_map_obj_control(L4_BASE_TASK_CAP, L4_MAP_ITEM_MAP)));

  if (err)
  {
    printf("Could not map task capability into task.\n");
    return 1;
  }

  err = l4_error(l4_task_map(task, L4RE_THIS_TASK_CAP,
                             l4_obj_fpage(thread, 0, L4_CAP_FPAGE_RWS),
                             l4_map_obj_control(L4_BASE_THREAD_CAP, L4_MAP_ITEM_MAP)));

  if (err)
  {
    printf("Could not map thread capability into task.\n");
    return 1;
  }

  err = l4_error(l4_task_map(task, L4RE_THIS_TASK_CAP,
                             l4_obj_fpage(l4re_env()->factory, 0, L4_CAP_FPAGE_RWS),
                             l4_map_obj_control(L4_BASE_FACTORY_CAP, L4_MAP_ITEM_MAP)));

  if (err)
  {
    printf("Could not map factory capability into task.\n");
    return 1;
  }

  err = l4_error(l4_task_map(task, L4RE_THIS_TASK_CAP,
                             l4_obj_fpage(l4re_env()->log, 0, L4_CAP_FPAGE_RWS),
                             l4_map_obj_control(L4_BASE_LOG_CAP, L4_MAP_ITEM_MAP)));

  if (err)
  {
    printf("Could not map log capability into task.\n");
    return 1;
  }

  err = l4_error(l4_task_map(task, L4RE_THIS_TASK_CAP,
                             l4_obj_fpage(l4re_env()->scheduler, 0, L4_CAP_FPAGE_RWS),
                             l4_map_obj_control(L4_BASE_SCHEDULER_CAP, L4_MAP_ITEM_MAP)));

  if (err)
  {
    printf("Could not map scheduler capability into task.\n");
    return 1;
  }

  /* Configure the thread with the region manager acting as pager and exception
     handler. The UTCB will be situated at an address supported by a dataspace
     attached to the new task. */

  printf("Configure thread...\n");

  l4_thread_control_start();
  l4_thread_control_pager(0x10 << L4_CAP_SHIFT);
  l4_thread_control_exc_handler(0x10 << L4_CAP_SHIFT);
  l4_thread_control_bind((l4_utcb_t *) utcb_start, task);
  err = l4_error(l4_thread_control_commit(thread));

  if (err)
  {
    printf("Could not configure thread.\n");
    return 1;
  }

  /* Start the new thread. */

  printf("Schedule thread...\n");

  err = l4_error(l4_thread_ex_regs(thread, 0x1000ae3, 0x2000000 /* stack top */, 0));

  if (err)
  {
    printf("Could not set thread registers.\n");
    return 1;
  }

  printf("Run thread...\n");

  l4_sched_param_t sp = l4_sched_param(L4RE_MAIN_THREAD_PRIO);

  err = l4_error(l4_scheduler_run_thread(l4re_env()->scheduler, thread, &sp));

  if (err)
  {
    printf("Could not run thread.\n");
    return 1;
  }

  printf("Finished.\n");
  while (1);

  return 0;
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
